package server.database.schema.app;

import communication.CqeType;

import exception.DatabaseException;
import exception.ParserException;
import exception.UnknownUserException;
import server.data.ConfidentialityPolicy;
import server.data.PriorUser;
import server.data.User;
import server.database.NewMaintenanceDatabase;
import server.database.sql.SQLDatabase;
import server.database.sql.SQLExpression;
import server.database.schema.app.dict.DiagnosisDictionary;
import server.database.schema.app.dict.NameDictionary;
import server.database.schema.app.dict.PrescriptionDictionary;
import server.database.schema.maintenance.MaintenanceDatabaseSchema;
import server.parser.Formula;

public class PreCQEKrankheitExampleLena extends MaintenanceDatabaseSchema{
	
	public final IllTableSchema ill;
	public final TreatTableSchema treat;
	public final NameDictionary nameDict;
	public final DiagnosisDictionary diagnosisDict;
	public final PrescriptionDictionary prescriptionDict;
	
	public PreCQEKrankheitExampleLena() throws DatabaseException {
		super("PreCQE Krankheit Example von Lena");
		
		this.ill = new IllTableSchema();
		this.treat = new TreatTableSchema();
		
		// Add all tables to one list so that it can be iterated in the correct order easily.
		this.appTables.add( this.ill );
		this.appTables.add( this.treat );
		
		
		this.nameDict = new NameDictionary();
		this.diagnosisDict = new DiagnosisDictionary();
		this.prescriptionDict = new PrescriptionDictionary();
		
		this.maintenanceTables.add( this.nameDict );
		this.maintenanceTables.add( this.diagnosisDict );
		this.maintenanceTables.add( this.prescriptionDict );
	}

	@Override
	public void fillAppDBWithContents(SQLDatabase db) throws DatabaseException {
		this.fillIll(db);
		this.fillTreat(db);
	}
	
	private void fillIll(SQLDatabase db) throws DatabaseException {
		SQLExpression sql;
		
		sql = db.sql();
		sql.set( IllTableSchema.NAME, "Lisa" );
		sql.set( IllTableSchema.DIAGNOSIS, "Flu" );
		db.insert(this.ill);
		
		sql = db.sql();
		sql.set( IllTableSchema.NAME, "Mary" );
		sql.set( IllTableSchema.DIAGNOSIS, "Aids" );
		db.insert(this.ill);
		
		sql = db.sql();
		sql.set( IllTableSchema.NAME, "Mary" );
		sql.set( IllTableSchema.DIAGNOSIS, "Flu" );
		db.insert(this.ill);
		
		sql = db.sql();
		sql.set( IllTableSchema.NAME, "Mary" );
		sql.set( IllTableSchema.DIAGNOSIS, "Sinusitis" );
		db.insert(this.ill);
	}
	
	private void fillTreat(SQLDatabase db) throws DatabaseException {
		SQLExpression sql;
		
		sql = db.sql();
		sql.set( TreatTableSchema.NAME, "Mary" );
		sql.set( TreatTableSchema.PRESCRIPTION, "MedicineA" );
		db.insert(this.treat);
		
		sql = db.sql();
		sql.set( TreatTableSchema.NAME, "Mary" );
		sql.set( TreatTableSchema.PRESCRIPTION, "MedicineB" );
		db.insert(this.treat);
		
		sql = db.sql();
		sql.set( TreatTableSchema.NAME, "Mary" );
		sql.set( TreatTableSchema.PRESCRIPTION, "MedicineC" );
		db.insert(this.treat);
		
		sql = db.sql();
		sql.set( TreatTableSchema.NAME, "Lisa" );
		sql.set( TreatTableSchema.PRESCRIPTION, "MedicineB" );
		db.insert(this.treat);
	}

	@Override
	public void fillMaintenanceDBWithContents(NewMaintenanceDatabase db) throws DatabaseException {
		super.createDefaultAccessRights(db);
		super.createAdminUser(db);
		int userId = super.createTestUser(db, 1, "PreCQE");
		try {
			this.fillPrior(db.getUserManagement().load(userId));
			this.fillConfidentialityPolicy(db.getUserManagement().load(userId));
		} catch (UnknownUserException e) {
			throw new DatabaseException("User does not exists, check your code.", e);
		}
		

		super.linkDictionary( db, this.nameDict, IllTableSchema.NAME );
		super.linkDictionary( db, this.diagnosisDict, IllTableSchema.DIAGNOSIS );
		super.linkDictionary( db, this.nameDict, TreatTableSchema.NAME );
		super.linkDictionary( db, this.prescriptionDict, TreatTableSchema.PRESCRIPTION );
	}
	
	public void fillPrior(User user) throws DatabaseException {
		try {
			PriorUser prior = user.getPriorUser();
			prior.add(new Formula("FORALL X( treat(X,\"MedicineC\") IMPL ill(X,\"Flu\") AND ill(X,\"Sinusitis\"))"));
			prior.add(new Formula("EXISTS X (treat(X,\"MedicineB\") AND ill(X,\"Flu\"))"));
			prior.add(new Formula("FORALL X (treat(X,\"MedicineA\") AND treat(X,\"MedB\") IMPL ill(X,\"Aids\"))"));
		} catch ( ParserException e ) {
			throw new DatabaseException("Formula in example wrong, check your code.", e);
		}
	}
	
	public void fillConfidentialityPolicy(User user) throws DatabaseException {
		try {
			ConfidentialityPolicy policy = user.getConfidentialityPolicy();
			policy.add(new Formula("EXISTS X ill(X,\"Aids\")"), CqeType.PolicyType.POTENTIAL_SECRETS, CqeType.PolicyPreservation.CONTINUOUS);
			policy.add(new Formula("ill(\"Lisa\",\"Flu\")"), CqeType.PolicyType.POTENTIAL_SECRETS, CqeType.PolicyPreservation.CONTINUOUS);
		} catch ( ParserException e ) {
			throw new DatabaseException("Formula in example wrong, check your code.", e);
		}
	}
}
